//=============================================================================
// KNH_MiniMenuWindow.js
//=============================================================================

/*:
 * @plugindesc ミニメニュー表示プラグイン
 * @author こんにちは
 *
 * @version 1.0.0 2017/10/27 公開
 *
 * @param アイコン画像
 * @desc アイコン画像を指定。
 * アイコンを必要数横に並べた画像を用意し、img/systemに保存。
 * @default 
 * @type file
 * @require 1
 * @dir img/system
 *
 * @param アイコンの大きさ
 * @desc 項目アイコンの一辺の長さ。(デフォルト:72)
 * @default 72
 * @type number
 *
 * @param 通常メニューの呼び出し禁止
 * @desc キャンセルによるメニューの呼び出しを禁止するか。
 * @default true
 * @type boolean
 * @on ON
 * @off OFF
 *
 * @param インデックス0 の実行内容
 * @desc インデックス0 のアイコンを押した時に実行する
 * コモンイベントのID。
 * @default 1
 * @type common_event
 *
 * @param インデックス1 の実行内容
 * @desc インデックス1 のアイコンを押した時に実行する
 * コモンイベントのID。
 * @default 1
 * @type common_event
 *
 * @param インデックス2 の実行内容
 * @desc インデックス2 のアイコンを押した時に実行する
 * コモンイベントのID。
 * @default 1
 * @type common_event
 *
 * @param インデックス3 の実行内容
 * @desc インデックス3 のアイコンを押した時に実行する
 * コモンイベントのID。
 * @default 1
 * @type common_event
 *
 * @param インデックス4 の実行内容
 * @desc インデックス4 のアイコンを押した時に実行する
 * コモンイベントのID。
 * @default 1
 * @type common_event
 *
 * @param インデックス5 の実行内容
 * @desc インデックス5 のアイコンを押した時に実行する
 * コモンイベントのID。
 * @default 1
 * @type common_event
 *
 * @param インデックス6 の実行内容
 * @desc インデックス6 のアイコンを押した時に実行する
 * コモンイベントのID。
 * @default 1
 * @type common_event
 *
 * @param インデックス7 の実行内容
 * @desc インデックス7 のアイコンを押した時に実行する
 * コモンイベントのID。
 * @default 1
 * @type common_event
 *
 *
 * @help ミニメニューを表示するプラグインです。
 * 画面上に表示されている矢印を押すと、ミニメニューがにゅっと出てきます。
 *
 * ミニメニューが表示されてる間も、普通に移動とかはできます。
 * 矢印とミニメニューはイベント実行中は自動的に非表示になります。
 *
 * ミニメニューにはアイコンが並び、クリックorタップすると
 * プラグインパラメータで指定したIDのコモンイベントが呼び出されます。
 *
 * メニュー項目に使うアイコン画像素材が必要となります。
 * img/system に保存し、プラグインパラメータで指定してください。
 * プラグインパラメータで指定した大きさ(デフォルトで48*48)のアイコン画像を
 * 必要数横に並べてください。左から順にインデックス 0,1,2,… となります。
 * 
 * なお、キーボード操作には非対応です。
 *
 * ●プラグインコマンド
 *  【】内を入力してください。「\v[1]」等使用できます。
 *
 * 【ミニメニュー非表示】  ミニメニューと矢印を非表示にします。
 * 【ミニメニュー表示】  ミニメニューと矢印を表示します。
 *
 * 【ミニメニュー項目選択不可 0】
 *   インデックス0 の項目を選択不可にします。
 *
 * 【ミニメニュー項目選択可 2】
 *   インデックス2 の項目を選択可にします。
 *
 * ●利用規約
 * 　特に無いです。
 */
(function() {
    'use strict';
    var pluginName = 'KNH_MiniMenuWindow';

    var getBoolean = function(text) {
        return (applyCharacters(text) || '').toUpperCase() === 'ON';
    };
    
    var getNumber = function(text, max, min) {
        text = applyCharacters(text);
        if (isNaN(text)) return NaN;
        return Math.max(Math.min(Number(text), max || Infinity), min || -Infinity);
    };
    
    var getArrayString = function(text) {
        var values = applyCharacters(text).split(',');
        for (var i = 0; i < values.length; i++) {
            values[i] = values[i].trim();
            if (!isNaN(values[i])) values[i] = Number(values[i]);
        }
        return values;
    };
    
    var applyCharacters = function(text) {
        if (text === null) text = '';
        text = '' + text;
        text = text.replace(/\\/g, '\x1b');
        text = text.replace(/\x1b\x1b/g, '\\');
        text = text.replace(/\x1bV\[(\d+)\]/gi, function() {
            return $gameVariables.value(parseInt(arguments[1], 10));
        }.bind(this));
        text = text.replace(/\x1bV\[(\d+)\]/gi, function() {
            return $gameVariables.value(parseInt(arguments[1], 10));
        }.bind(this));
        text = text.replace(/\x1bN\[(\d+)\]/gi, function() {
            var actor = parseInt(arguments[1], 10) >= 1 ? $gameActors.actor(parseInt(arguments[1], 10)) : null;
            return actor ? actor.name() : '';
        }.bind(this));
        text = text.replace(/\x1bP\[(\d+)\]/gi, function() {
            var actor = parseInt(arguments[1], 10) >= 1 ? $gameParty.members()[parseInt(arguments[1], 10) - 1] : null;
            return actor ? actor.name() : '';
        }.bind(this));
        if ($dataSystem) text = text.replace(/\x1bG/gi, TextManager.currencyUnit);
        return text;
    };
    
    //-----------------------------------------------------------------------------
    // プラグインパラメータの取得
    //-----------------------------------------------------------------------------
    
    var paramText = function(paramName) {
        return applyCharacters(PluginManager.parameters(pluginName)[paramName]);
    };

    var paramBoolean = function(paramName) {
        return getBoolean(paramText(paramName));
    };
    
    var paramNumber = function(paramName, max, min) {
        return getNumber(paramText(paramName), max, min);
    };
    
    var paramArrayString = function(paramName) {
        return getArrayString(paramText(paramName));
    };

    var KNH_normalMenuEnabled = !paramBoolean('通常メニューの呼び出し禁止');
    var KNH_iconSize = paramNumber('アイコンの大きさ');
    var KNH_commonEvents = [];
    KNH_commonEvents.push(paramNumber('インデックス0 の実行内容'));
    KNH_commonEvents.push(paramNumber('インデックス1 の実行内容'));
    KNH_commonEvents.push(paramNumber('インデックス2 の実行内容'));
    KNH_commonEvents.push(paramNumber('インデックス3 の実行内容'));
    KNH_commonEvents.push(paramNumber('インデックス4 の実行内容'));
    KNH_commonEvents.push(paramNumber('インデックス5 の実行内容'));
    KNH_commonEvents.push(paramNumber('インデックス6 の実行内容'));
    KNH_commonEvents.push(paramNumber('インデックス7 の実行内容'));
    
    //-----------------------------------------------------------------------------
    // プラグインコマンドの追加
    //-----------------------------------------------------------------------------
    
    var _Game_Interpreter_pluginCommand      = Game_Interpreter.prototype.pluginCommand;
    Game_Interpreter.prototype.pluginCommand = function(command, args) {
        _Game_Interpreter_pluginCommand.call(this, command, args);
        try {
            this.pluginCommandMiniMenuWindow(command, args);
        } catch (e) {
            if ($gameTemp.isPlaytest() && Utils.isNwjs()) {
                var window = require('nw.gui').Window.get();
                if (!window.isDevToolsOpen()) {
                    var devTool = window.showDevTools();
                    devTool.moveTo(0, 0);
                    devTool.resizeTo(Graphics.width, Graphics.height);
                    window.focus();
                }
            }
            console.log('プラグインコマンドの実行中にエラーが発生しました。');
            console.log('- コマンド名 　: ' + command);
            console.log('- コマンド引数 : ' + args);
            console.log('- エラー原因   : ' + e.toString());
        }
    };
    
    Game_Interpreter.prototype.pluginCommandMiniMenuWindow = function(command, args) {
        var text;
        switch (command) {
            case 'ミニメニュー非表示':
                $gameSystem.miniMenu().noVisible = true;
                break;
            case 'ミニメニュー表示':
                $gameSystem.miniMenu().noVisible = false;
                break;
            case 'ミニメニュー項目選択不可':
                $gameSystem.miniMenu().commandCannot[getNumber(args[0])] = true;
            break;
            case 'ミニメニュー項目選択可':
                $gameSystem.miniMenu().commandCannot[getNumber(args[0])] = false;
            break;
        }
    };
    
    //-----------------------------------------------------------------------------
    // Game_System
    //-----------------------------------------------------------------------------
    
    Game_System.prototype.miniMenu = function() {
        if (!this._miniMenu) this._miniMenu = {noVisible : false, commandCannot : []};
        return this._miniMenu;
    };
    
    Game_System.prototype.isMenuEnabled = function() {
        return this._menuEnabled && KNH_normalMenuEnabled;
    };
    
    //-----------------------------------------------------------------------------
    // Scene_Map
    //-----------------------------------------------------------------------------
    
    var _Scene_Map_initialize = Scene_Map.prototype.initialize;
    Scene_Map.prototype.initialize = function() {
        _Scene_Map_initialize.apply(this, arguments);
    };
    
    var _Scene_Map_start = Scene_Map.prototype.start;
    Scene_Map.prototype.start = function() {
        _Scene_Map_start.apply(this, arguments);
        this._miniMenuWindow.refresh();
    };

    var _Scene_Map_terminate = Scene_Map.prototype.terminate;
    Scene_Map.prototype.terminate = function() {
        _Scene_Map_terminate.apply(this, arguments);
        this.removeChild(this._miniMenuWindow);
    };
    
    var _Scene_Map_createDisplayObjects = Scene_Map.prototype.createDisplayObjects;
    Scene_Map.prototype.createDisplayObjects = function() {
        _Scene_Map_createDisplayObjects.apply(this, arguments);
        this.createMiniMenuWindow();
    };
    
    Scene_Map.prototype.createMiniMenuWindow = function() {
        this._miniMenuWindow = new Window_MiniMenu();
        this.addChild(this._miniMenuWindow);
    };
    
    var _Scene_Map_processMapTouch = Scene_Map.prototype.processMapTouch;
    Scene_Map.prototype.processMapTouch = function() {
        if (!this._miniMenuWindow.isTouchedInsideFrame()) _Scene_Map_processMapTouch.apply(this, arguments);
    };
    
    var _Scene_Map_update = Scene_Map.prototype.update;
    Scene_Map.prototype.update = function() {
        _Scene_Map_update.apply(this, arguments);
        this.updateMiniMenu();
    };
    
    Scene_Map.prototype.updateMiniMenu = function() {
        var mw = this._miniMenuWindow;
        if ($gameSystem.miniMenu().noVisible || $gameMap.isEventRunning() || this.menuCalling) mw._positionType = 0;
        else if (mw._positionType === 0) mw._positionType = $gameSystem._saveMiniMenuPositionType || 1;
        if (this.menuCalling) mw.x = Graphics.boxWidth;
        if (mw._hideCount > 0) mw.x = Graphics.boxWidth, mw._hideCount--;
    };
    
    //-----------------------------------------------------------------------------
    // Window_MiniMenu
    //-----------------------------------------------------------------------------
    
    function Window_MiniMenu() {
        this.initialize.apply(this, arguments);
    }

    Window_MiniMenu.prototype = Object.create(Window_Selectable.prototype);
    Window_MiniMenu.prototype.constructor = Window_MiniMenu;

    Window_MiniMenu.prototype.initialize = function() {
        var x = Graphics.boxWidth;
        var y = 12;
        var height = this.windowHeight();
        Window_Selectable.prototype.initialize.call(this, x, y, 0, height);
        this._commonEvents = [];
        this._positionType = this._beforePositionType = $gameSystem._saveMiniMenuPositionType || 1;
        this._hideCount = 0;
        this._activeWait = 0;
        this._iconBitmap = ImageManager.loadSystem('KNH_MenuIcon');
        this.setHandler('ok', this.processOk.bind(this));
    };
    
    Window_MiniMenu.prototype.standardPadding = function() {
        return 8;
    };
    
    Window_MiniMenu.prototype.spacing = function() {
        return 0;
    };
    
    Window_MiniMenu.prototype.windowHeight = function() {
        return this.itemIconSize() + this.padding * 2 + 16;
    };

    Window_MiniMenu.prototype.windowWidth = function() {
        return this._commonEvents.length * this.itemIconSize() + 72;
    };
    
    Window_MiniMenu.prototype.itemWidth = function() {
        return this.itemIconSize();
    };

    Window_MiniMenu.prototype.itemHeight = function() {
        return this.itemIconSize();
    };
    
    Window_MiniMenu.prototype.itemIconSize = function() {
        return KNH_iconSize;
    };
    
    Window_MiniMenu.prototype.maxCols = function() {
        return 8;
    };

    Window_MiniMenu.prototype.maxItems = function() {
        var max = Math.floor((Graphics.boxWidth - this.padding * 2 - 48) / this.itemWidth());
        return this._commonEvents ? Math.min(this._commonEvents.length + 1, max) : 1;
    };

    Window_MiniMenu.prototype.isEnabled = function(index) {
        return !$gameSystem.miniMenu().commandCannot[index];
    };
    
    Window_MiniMenu.prototype.itemRect = function(index) {
        var rect;
        var arrowWidth = 48;
        if (index === 0) {
            rect = new Rectangle();
            rect.width = arrowWidth;
            rect.height = this.height - this.padding * 2;
            rect.x = rect.y = 0;
        } else {
            rect = Window_Selectable.prototype.itemRect.apply(this, arguments);
            rect.x -= this.itemIconSize() - arrowWidth;
            rect.y += 8;
        }
        return rect;
    };
    
    Window_MiniMenu.prototype.makeItemList = function() {
        var bw = this._iconBitmap.canvas.width;
        var listLength = (bw - bw % this.itemIconSize()) / this.itemIconSize();
        this._commonEvents = KNH_commonEvents.slice(0, listLength);
    };

    Window_MiniMenu.prototype.refresh = function() {
        this.makeItemList();
        this.width = this.windowWidth();
        this.height = this.windowHeight();
        this.createContents();
        this.drawAllItems();
    };

    Window_MiniMenu.prototype.drawItem = function(index) {
        var rect = this.itemRect(index);
        if (index > 0) {
            this.changePaintOpacity(this.isEnabled(index - 1));
            this.drawMiniMenuIcon(index - 1, rect.x, rect.y);
            this.changePaintOpacity(true);
        } else {
            this.contents.fontSize = 60;
            var text = this._positionType === 1 ? '<' : '>';
            this.drawText(text, rect.x, rect.y + rect.height / 2 - 20, rect.width - 8, 'center');
            this.resetFontSettings();
        }
    };
    
    Window_MiniMenu.prototype.drawMiniMenuIcon = function(index, x, y) {
        var bitmap = this._iconBitmap;
        var pw = this.itemIconSize();
        var ph = this.itemIconSize();
        var sx = index * this.itemIconSize();
        this.contents.blt(bitmap, sx, 0, pw, ph, x, y);
    };
    
    Window_MiniMenu.prototype.processCursorMove = function() {
    };
    
    Window_MiniMenu.prototype.processOk = function() {
        if (this._index === 0) {
            this._positionType === 1 ? SoundManager.playOk() : SoundManager.playCancel();
            this._positionType = $gameSystem._saveMiniMenuPositionType = this._positionType % 2 + 1;
        } else {
            if (this.isEnabled(this._index - 1)) {
                SoundManager.playOk();
                var ceId = KNH_commonEvents[this._index - 1];
                $gameMap._interpreter.setup($dataCommonEvents[ceId].list);
                this.x = Graphics.boxWidth;
                this._hideCount = 6;
            } else {
                SoundManager.playBuzzer();
            }
        }
        this._activeWait = 6;
    };
    
    Window_MiniMenu.prototype.update = function() {
        Window_Selectable.prototype.update.apply(this, arguments);
        if (this._positionType !== this._beforePositionType) {
            this._beforePositionType = this._positionType;
            this.refresh();
        }
        var gw = Graphics.boxWidth;
        var targetX = [gw, gw - 48, gw - this.width + 12][this._positionType] || 0;
        targetX > 0 ? this.activate() : this.deactivate();
        this.x += (targetX - this.x) / 3;
        if (this._activeWait-- <= 0) {
            var x  = this.canvasToLocalX(TouchInput.x);
            var y  = this.canvasToLocalY(TouchInput.y);
            var hitIndex = this.hitTest(x, y);
            this.select(hitIndex);
            if (hitIndex >= 0 && TouchInput.isTriggered() && this.isTouchOkEnabled()) this.processOk();
        }
    };
    
    var _Window_MiniMenu_isCursorVisible = Window_MiniMenu.prototype.isCursorVisible;
    Window_MiniMenu.prototype.isCursorVisible = function(triggered) {
        if (Utils.isMobileDevice()) return false;
        return _Window_MiniMenu_isCursorVisible.apply(this, arguments);
    };
    
    Window_MiniMenu.prototype.processHandling = function() {};
    
    Window_MiniMenu.prototype.onTouch = function(triggered) {};
    
    //-----------------------------------------------------------------------------
    // TouchInput  マウス位置の記録を常に行う
    //-----------------------------------------------------------------------------
    
    TouchInput._onMouseMove = function(event) {
        var x = Graphics.pageToCanvasX(event.pageX);
        var y = Graphics.pageToCanvasY(event.pageY);
        this._onMove(x, y);
    };
    
})();
