//=============================================================================
// KNH_TouchMoveStraight.js
//=============================================================================

/*:
 * @plugindesc タッチ移動をまっすぐにするプラグイン
 * @author こんにちは
 *
 * @version 1.0.0 2017/10/27 公開
 *
 * @param 移動フレーム数
 * @desc 移動にかかる時間をフレーム数で指定。
 * @default 6
 * @type number
 *
 * @help マップをタッチしたとき、途中の壁やイベントなどを無視し
 * タッチした座標へまっすぐ一直線に移動するようになります。
 * 「プレイヤーをカーソル化したいシーン」に使うといいと思います。
 *
 * ●仕組み
 * 一直線に進む移動は「高さ0のジャンプ」によって再現しています。
 * なので壁もイベントもすり抜けます。
 * なお、通常のジャンプの仕様は変わらないので安心してください。
 *
 * ●プラグインコマンド
 *  【】内を入力してください。
 *
 * 【タッチ移動まっすぐ化】     タッチ移動がまっすぐになります
 * 【タッチ移動まっすぐ化解除】  タッチ移動を普通にもどします
 *
 * ●利用規約
 *   特に無いです。
 */

(function() {
    'use strict';
    var pluginName = 'KNH_TouchMoveStraight';
    
    var getNumber = function(text, max, min) {
        text = applyCharacters(text);
        if (isNaN(text)) return NaN;
        return Math.max(Math.min(Number(text), max || Infinity), min || -Infinity);
    };
    
    var applyCharacters = function(text) {
        if (text === null) text = '';
        text = '' + text;
        text = text.replace(/\\/g, '\x1b');
        text = text.replace(/\x1b\x1b/g, '\\');
        text = text.replace(/\x1bV\[(\d+)\]/gi, function() {
            return $gameVariables.value(parseInt(arguments[1], 10));
        }.bind(this));
        text = text.replace(/\x1bV\[(\d+)\]/gi, function() {
            return $gameVariables.value(parseInt(arguments[1], 10));
        }.bind(this));
        text = text.replace(/\x1bN\[(\d+)\]/gi, function() {
            var actor = parseInt(arguments[1], 10) >= 1 ? $gameActors.actor(parseInt(arguments[1], 10)) : null;
            return actor ? actor.name() : '';
        }.bind(this));
        text = text.replace(/\x1bP\[(\d+)\]/gi, function() {
            var actor = parseInt(arguments[1], 10) >= 1 ? $gameParty.members()[parseInt(arguments[1], 10) - 1] : null;
            return actor ? actor.name() : '';
        }.bind(this));
        if ($dataSystem) text = text.replace(/\x1bG/gi, TextManager.currencyUnit);
        return text;
    };
    
    //-----------------------------------------------------------------------------
    // プラグインパラメータの取得
    //-----------------------------------------------------------------------------
    
    var param = function(paramName) {
        return PluginManager.parameters(pluginName)[paramName];
    };
    
    var paramText = function(paramName) {
        return applyCharacters(PluginManager.parameters(pluginName)[paramName]);
    };
    
    var paramNumber = function(paramName, max, min) {
        return getNumber(paramText(paramName), max, min);
    };

    var KNH_straightMoveFrame = paramNumber('移動フレーム数');
    
    //-----------------------------------------------------------------------------
    // プラグインコマンドの追加
    //-----------------------------------------------------------------------------
    
    var _Game_Interpreter_pluginCommand      = Game_Interpreter.prototype.pluginCommand;
    Game_Interpreter.prototype.pluginCommand = function(command, args) {
        _Game_Interpreter_pluginCommand.call(this, command, args);
        try {
            this.pluginCommandTouchMoveStraight(command, args);
        } catch (e) {
            if ($gameTemp.isPlaytest() && Utils.isNwjs()) {
                var window = require('nw.gui').Window.get();
                if (!window.isDevToolsOpen()) {
                    var devTool = window.showDevTools();
                    devTool.moveTo(0, 0);
                    devTool.resizeTo(Graphics.width, Graphics.height);
                    window.focus();
                }
            }
            console.log('プラグインコマンドの実行中にエラーが発生しました。');
            console.log('- コマンド名 　: ' + command);
            console.log('- コマンド引数 : ' + args);
            console.log('- エラー原因   : ' + e.toString());
        }
    };
    
    Game_Interpreter.prototype.pluginCommandTouchMoveStraight = function(command, args) {
        var text;
        switch (command) {
            case 'タッチ移動まっすぐ化':
                $gameSystem.touchMoveStraight().moveStraight = true;
                break;
            case 'タッチ移動まっすぐ化解除':
                $gameSystem.touchMoveStraight().moveStraight = false;
                break;
        }
    };

    //-----------------------------------------------------------------------------
    // Game_System
    //-----------------------------------------------------------------------------
    
    Game_System.prototype.touchMoveStraight = function() {
        if (!this._touchMoveStraight) this._touchMoveStraight = {moveStraight : false};
        return this._touchMoveStraight;
    };
    
    //-----------------------------------------------------------------------------
    // Game_Player
    //-----------------------------------------------------------------------------
    
    var _Game_Player_moveByInput = Game_Player.prototype.moveByInput;
    Game_Player.prototype.moveByInput = function() {
        if ($gameSystem.touchMoveStraight().moveStraight && this.getInputDirection() === 0) { //フラグON && タッチ移動
            if (!this.isMoving() && this.canMove()) {
                if ($gameTemp.isDestinationValid()){
                    var x = $gameTemp.destinationX();
                    var y = $gameTemp.destinationY();
                    this._jumpStraightFlag = true;
                    this.jump(x - this._x, y - this._y);
                }
            }
        } else {
            _Game_Player_moveByInput.apply(this, arguments);
        }
    };
    
    var _Game_Player_jump = Game_Player.prototype.jump;
    Game_Player.prototype.jump = function(xPlus, yPlus) {
        _Game_Player_jump.call(this, xPlus, yPlus);
        if (this._jumpStraightFlag) {
            this._jumpStraightFlag = false;
            this._jumpStraight = true;
            this._jumpCount = KNH_straightMoveFrame;
            this._jumpPeak = this._jumpCount / 2;
        } else {
            this._jumpStraight = false;
        }
    };
    
    var _Game_Player_jumpHeight = Game_Player.prototype.jumpHeight;
    Game_Player.prototype.jumpHeight = function() {
        if (!this.isJumping() || this._jumpStraight) return 0;
        else return _Game_Player_jumpHeight.call(this);
    };
    
})();
