//=============================================================================
// KNH_Effect.js
//=============================================================================

/*:
 * @plugindesc エフェクト表示プラグイン
 * @author こんにちは
 *
 * @version 1.0.0 2017/10/27 公開
 *
 * @help アニメをいい感じにランダムで表示し続けます
 * キャラで関数 setEffect(アニメID, 頻度, 時間) を実行してください
 * ・頻度：だいたい1秒あたりの個数
 * ・時間：指定すると時間で消える 単位はフレーム
 *
 * ●利用規約
 * 　特に無いです。
 */
(function() {
    'use strict';
    var pluginName = 'KNH_Effect';

    var getBoolean = function(text) {
        return (applyCharacters(text) || '').toUpperCase() === 'ON';
    };
    
    var getNumber = function(text, max, min) {
        text = applyCharacters(text);
        if (isNaN(text)) return NaN;
        return Math.max(Math.min(Number(text), max || Infinity), min || -Infinity);
    };
    
    var getArrayString = function(text) {
        var values = applyCharacters(text).split(',');
        for (var i = 0; i < values.length; i++) {
            values[i] = values[i].trim();
            if (!isNaN(values[i])) values[i] = Number(values[i]);
        }
        return values;
    };
    
    var applyCharacters = function(text) {
        if (text === null) text = '';
        text = '' + text;
        text = text.replace(/\\/g, '\x1b');
        text = text.replace(/\x1b\x1b/g, '\\');
        text = text.replace(/\x1bV\[(\d+)\]/gi, function() {
            return $gameVariables.value(parseInt(arguments[1], 10));
        }.bind(this));
        text = text.replace(/\x1bV\[(\d+)\]/gi, function() {
            return $gameVariables.value(parseInt(arguments[1], 10));
        }.bind(this));
        text = text.replace(/\x1bN\[(\d+)\]/gi, function() {
            var actor = parseInt(arguments[1], 10) >= 1 ? $gameActors.actor(parseInt(arguments[1], 10)) : null;
            return actor ? actor.name() : '';
        }.bind(this));
        text = text.replace(/\x1bP\[(\d+)\]/gi, function() {
            var actor = parseInt(arguments[1], 10) >= 1 ? $gameParty.members()[parseInt(arguments[1], 10) - 1] : null;
            return actor ? actor.name() : '';
        }.bind(this));
        if ($dataSystem) text = text.replace(/\x1bG/gi, TextManager.currencyUnit);
        return text;
    };
    
    //-----------------------------------------------------------------------------
    // プラグインパラメータの取得
    //-----------------------------------------------------------------------------
    
    var paramText = function(paramName) {
        return applyCharacters(PluginManager.parameters(pluginName)[paramName]);
    };

    var paramBoolean = function(paramName) {
        return getBoolean(paramText(paramName));
    };
    
    var paramNumber = function(paramName, max, min) {
        return getNumber(paramText(paramName), max, min);
    };
    
    var paramArrayString = function(paramName) {
        return getArrayString(paramText(paramName));
    };
    
    //-----------------------------------------------------------------------------
    // Game_CharacterBase
    //-----------------------------------------------------------------------------
    
    Game_CharacterBase.prototype.setEffect = function(id, frequency, time) {
        this._KNHEffect = {id : id || 0, frequency : frequency || 1, time : time || 0};
    };
    
    Game_CharacterBase.prototype.endEffect = function() {
        if (this._KNHEffect) this._KNHEffect.id = 0;
    };
    
    //-----------------------------------------------------------------------------
    // Sprite_Character
    //-----------------------------------------------------------------------------
    
    var _Sprite_Character_updateOther = Sprite_Character.prototype.updateOther;
    Sprite_Character.prototype.updateOther = function() {
        _Sprite_Character_updateOther.apply(this, arguments);
        this._KNHEffect = this._character._KNHEffect
    };
    
    //-----------------------------------------------------------------------------
    // Sprite_Base
    //-----------------------------------------------------------------------------
    
    var _Sprite_Base_update = Sprite_Base.prototype.update;
    Sprite_Base.prototype.update = function() {
        _Sprite_Base_update.apply(this, arguments);
        if (this._KNHEffect && this._KNHEffect.id) {
            this._KNHEffectCount = this._KNHEffectCount || (Math.randomInt(100)) / 100;
            if (this._KNHEffectCount > 1) {
                this._KNHEffectFlag = true;
                this.startAnimation($dataAnimations[this._KNHEffect.id], !!Math.randomInt(2), 0);
                this._KNHEffectCount--;
            }
            this._KNHEffectCount += this._KNHEffect.frequency / 60 * (Math.randomInt(201)) / 100;
            if (--this._KNHEffect.time === 0) this._KNHEffect.id = 0;
        }
    };
    
    Sprite_Base.prototype.startAnimation = function(animation, mirror, delay) {
        var sprite = new Sprite_Animation();
        if (this._KNHEffectFlag) {
            sprite._isKNHEffect = true;
            this._KNHEffectFlag = false;
        }
        sprite.setup(this._effectTarget, animation, mirror, delay);
        this.parent.addChild(sprite);
        this._animationSprites.push(sprite);
    };
    
    //-----------------------------------------------------------------------------
    // Sprite_Animation
    //-----------------------------------------------------------------------------
    
    var _Sprite_Animation_updateCellSprite = Sprite_Animation.prototype.updateCellSprite
    Sprite_Animation.prototype.updateCellSprite = function(sprite, cell) {
        _Sprite_Animation_updateCellSprite.apply(this, arguments);
        if (this._isKNHEffect) {
            var ax = (this._target._character && this._target._character._eAx) ? this._target._character._eAx : 0;
            var ay = (this._target._character && this._target._character._eAy) ? this._target._character._eAy : 0;
            var aw = (this._target._character && this._target._character._eAw) ? this._target._character._eAw : 1;
            var ah = (this._target._character && this._target._character._eAh) ? this._target._character._eAh : 1;
            if (sprite._KNHposInfo === undefined)
                sprite._KNHposInfo = {
                    x     : aw * this._target.width  * (-60 + Math.randomInt(121)) / 100 + ax,
                    y     : ah * this._target.height * (-60 + Math.randomInt(121)) / 100 + ay,
                    scale : (20 + Math.randomInt(81)) / 100
                };
            sprite.x = sprite._KNHposInfo.x;
            sprite.y = sprite._KNHposInfo.y;
            sprite.scale.x = sprite.scale.y = sprite._KNHposInfo.scale;
        }
    };
    
    var _Sprite_Animation_setupRate = Sprite_Animation.prototype.setupRate
    Sprite_Animation.prototype.setupRate = function() {
        this._isKNHEffect ? (this._rate = 3) : _Sprite_Animation_setupRate;
    };
    
})();
